/*
 *	(c) 1998-99 Sassafras Software Inc.
 *
 *	Thread events.  One thread waits on an event, another
 *	thread signals the event to wake the first thread up.
 */

/**********************************************************************

  YAAF is Copyright 1997-1999 by William Woody.

  Redistribution and use in source and binary forms, with or without 
  modification, are permitted provided that the following conditions 
  are met:

  Redistributions of source code must retain the above copyright notice, 
  this list of conditions and the following disclaimer.

  Neither the name of the copyright holder or other contributors may
  be used to promote products derived from this software without
  specific written prior permission.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, 
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES 
  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND 
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS 
  BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN 
  ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN 
  CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE 
  SOFTWARE.
  
  Contact:   woody@alumni.caltech.edu
  Web page:  http://www.pandawave.com/yaaf/

**********************************************************************/

#ifndef SS_TTHREADEVENT
#define SS_TTHREADEVENT

#include <XConfig.h>
#include <XThread.h>

/************************************************************************/
/*																		*/
/*	Class Declaration													*/
/*																		*/
/************************************************************************/

/*$	XGThreadEvent
 *
 *		This provides a simple "thread event" object. This permits to
 *	send simple events from one thread to another.
 */

class XGThreadEvent {
	public:
						XGThreadEvent (bool manual = false, bool state = false);
		virtual			~XGThreadEvent (void);

		void			Set (void);
		void			Reset (void);
		bool			Test (void);
		bool			Wait (unsigned long timeout);

	private:

#if (OPT_MACOS == 1) || (OPT_XWIN == 1)
		XGSemaphore		fSemaphore;		/* this is a spin-loop on the Mac */
		bool			fState;
		bool			fManual;
#endif

#if OPT_WINOS == 1
		HANDLE			fHandle;		/* use native events on Windows */
#endif
};

#endif SS_TTHREADEVENT
